clc
close all
clear all

global mdl
global xdata
global ydata

warning("off")
dt = datetime('now','TimeZone','local','Format','y-MMM-d HH_mm_ss');
dtstr = string(dt);
filename = strcat(pwd,'\data2.xlsx');
delete("data2_*.txt");
outFile = strcat("data2_",dtstr,".txt");
diary(outFile)

fprintf("Program  clem2\n\n");
fprintf("%s\n", dtstr);
fprintf("Please wait ...\n\n");
fprintf("Excel file used is %s\n", filename);
fprintf("Diary file used is %s\n", outFile);
xyData = readmatrix(filename,'Sheet','Data');

ydata = xyData(:,1);
xdata = xyData(:,2:end);
[maxrows,maxcols] = size(xdata);
xyTransf = readmatrix(filename,'Sheet','Transf');
xyTransf = xyTransf(:,2:end);

Lb = xyTransf(1,:);
Ub = xyTransf(2,:);
nVars = length(Lb);
iFrom = Lb;
iTo = Ub;
iStep = 1 + zeros(1,nVars);
idx = Lb;
bStop = false;
iter = 0;
fprintf("------------- Calculating Max Iters ----------------\n");
while ~bStop 
  iter = iter +1;
  % Start implementing the quasi–nested loops
  idx(1) = idx(1) + iStep(1);
  if idx(1) > iTo(1)
    idx(1) = iFrom(1);
    for i = 2:nVars
      idx(i) = idx(i) + iStep(i);
      if idx(i) > iTo(i) && i < nVars
        idx(i) = iFrom(i);
      elseif idx(i) > iTo(i) && i == nVars
        bStop = true;
      else
        break
      end
    end
  end
end
str = num2str(iter);
str = fliplr(regexprep(fliplr(str), '(\d+\.)?(\d{3})(?=\S+)', '$1$2,'));
fprintf("Max iters = %s\n", str);
MaxIters = iter;
FivePercent = fix(iter/20);

idx = Lb;
iter = 0;

bestX = zeros(1,length(Lb));
bestFx = 0;
% --------------------------------- start main loop
bStop = false;
iter = 0;
fprintf("------------- Maximizing Fx ----------------\n");
while ~bStop 
  iter = iter +1;
  if mod(iter, FivePercent) == 0
    percent = 100*iter/MaxIters;
    fprintf("Progress %5.2f%%\n", percent);
  end

  r2adj = myFit(idx);
  if r2adj > bestFx
    bestFx = r2adj;
    bestX = idx;
    bestMdl = mdl;
    fprintf("Iter %d, Fx = %e, X=[", iter, bestFx)
    fprintf("%d, ", bestX);
    fprintf("]\n");
  end
  % Start implementing the quasi–nested loops
  idx(1) = idx(1) + iStep(1);
  if idx(1) > iTo(1)
    idx(1) = iFrom(1);
    for i = 2:nVars
      idx(i) = idx(i) + iStep(i);
      if idx(i) > iTo(i) && i < nVars
        idx(i) = iFrom(i);
      elseif idx(i) > iTo(i) && i == nVars
        bStop = true;
      else
        break
      end
    end
  end
end
fprintf("Regression model\n")
format long
AdjR2 = bestMdl.Rsquared.Adjusted;
fprintf("Adjusted R-square = %14.12f\n", AdjR2);
mdl = bestMdl
bestX = [AdjR2 bestX];

res = zeros(2,length(bestX));
res(1,:) = bestX;
res(2,2) = mdl.Coefficients{1,1}';
res(2,3) = mdl.Coefficients{2,1}';
res(2,4) = mdl.Coefficients{3,1}';
writematrix(res(1,:), filename, 'Sheet', 'Results', 'Range', 'A2:Z2');
writematrix(res(2,2:4), filename, 'Sheet', 'Results', 'Range', 'B3:Z3');
 
% now do the % projections
ycalc = project(res);
percErr = 100.*(ycalc-ydata)./ydata;
ProjMat = [ydata xdata ycalc percErr];
writematrix(ProjMat, filename, 'Sheet', 'Project', 'Range', strcat("A2:Z", num2str(1+length(ydata))));


i1 = res(1,2);
i2 = res(1,3);
px1 = res(1,4);
px2 = res(1,5);
smodel = sprintf("Y = (%e + (%e)*", res(2,2), res(2,3));

if px1==0
  sX1 = sprintf("log(X%d)", i1);
elseif px1==4
  sX1 = sprintf("sqrt(X%d)", i1);
elseif px1==1
  sX1 = sprintf("X%d", i1);
elseif px1>1
  sX1 = sprintf("X%d^%d", i1, px1);
elseif px1<0
  sX1 = sprintf("1/X%d^%d", i1, abs(px1));
else
  sX1 = sprintf("X%d", i1);
end

if px2==0
  sX2 = sprintf("log(X%d)", i2);
elseif px2==4
  sX2 = sprintf("sqrt(X%d)", i2);
elseif px2==1
  sX2 = sprintf("X%d", i2);
elseif px2>1
  sX2 = sprintf("X%d^%d", i2, px2);
elseif px2<0
  sX2 = sprintf("1/X%d^%d", i2, abs(px2));
else
  sX2 = sprintf("X%d", i2);
end  

s = sprintf("%s)/(1 + (%e)* %s %s", sX1, res(2,4), sX2);
smodel = strcat(smodel, s, ")");
fprintf("%s\n", smodel)

format short
fprintf("Done!\n");
diary off
beep on
for i=1:3
  beep;
  pause(1);
end

function x = myfx(x,p)
  if p==0
    x = log(x);
  elseif p==4
    x = sqrt(x);  
  elseif p > 1
    x = x.^p;
  elseif p < 0
    x = 1./x.^abs(p);
  end
end

function r = myFit(coeff)
  global mdl
  global xdata
  global ydata


  y = ydata;
  i1 = coeff(1);
  i2 = coeff(2);
  px1 = coeff(3);
  px2 = coeff(4);

  x1 = myfx(xdata(:,i1),px1);
  x2 = -ydata.*myfx(xdata(:,i2),px2);
  X=[x1 x2];
  mdl = fitlm(X,y);
  r = mdl.Rsquared.Adjusted;
end

function ycalc = project(res)
  global xdata

  i1 = res(1,2);
  i2 = res(1,3);
  px1 = res(1,4);
  px2 = res(1,5);

  x1 = myfx(xdata(:,i1),px1); 
  x2 = myfx(xdata(:,i2),px2); 
  ycalc = (res(2,2) + res(2,3) * x1)./(1 + res(2,4)*x2);
end