clc
close all
clear all

global mdl
global xdata
global ydata

warning("off")
dt = datetime('now','TimeZone','local','Format','y-MMM-d HH_mm_ss');
dtstr = string(dt);
filename = strcat(pwd,'\data4.xlsx');
delete("data4_*.txt");
outFile = strcat("data4_",dtstr,".txt");
diary(outFile)

fprintf("Program  clem4\n\n");
fprintf("%s\n", dtstr);
fprintf("Please wait ...\n\n");
fprintf("Excel file used is %s\n", filename);
fprintf("Diary file used is %s\n", outFile);
xyData = readmatrix(filename,'Sheet','Data');

ydata = xyData(:,1);
xdata = xyData(:,2:end);
[maxrows,maxcols] = size(xdata);
xyTransf = readmatrix(filename,'Sheet','Transf');
xyTransf = xyTransf(:,2:end);

Lb = xyTransf(1,:);
Ub = xyTransf(2,:);
nVars = length(Lb);
iFrom = Lb;
iTo = Ub;
iStep = 1 + zeros(1,nVars);
idx = Lb;
bStop = false;
iter = 0;
fprintf("------------- Calculating Max Iters ----------------\n");
while ~bStop 
  iter = iter +1;
  % Start implementing the quasi–nested loops
  idx(1) = idx(1) + iStep(1);
  if idx(1) > iTo(1)
    idx(1) = iFrom(1);
    for i = 2:nVars
      idx(i) = idx(i) + iStep(i);
      if idx(i) > iTo(i) && i < nVars
        idx(i) = iFrom(i);
      elseif idx(i) > iTo(i) && i == nVars
        bStop = true;
      else
        break
      end
    end
  end
end
str = num2str(iter);
str = fliplr(regexprep(fliplr(str), '(\d+\.)?(\d{3})(?=\S+)', '$1$2,'));
fprintf("Max iters = %s\n", str);
MaxIters = iter;
FivePercent = fix(iter/20);

iFrom = Lb;
iTo = Ub;
iStep = 1 + zeros(1,nVars);
idx = Lb;
iter = 0;

bestX = zeros(1,length(Lb));
bestFx = 0;
% --------------------------------- start main loop
bStop = false;
iter = 0;
fprintf("------------- Maximizing Fx ----------------\n");
while ~bStop 
  iter = iter +1;
  if mod(iter, FivePercent) == 0
    percent = 100*iter/MaxIters;
    fprintf("Progress %5.2f%%\n", percent);
  end

  r2adj = myFit(idx);
  if r2adj > bestFx
    bestFx = r2adj;
    bestX = idx;
    bestMdl = mdl;
    fprintf("Iter %d, Fx = %e, X=[", iter, bestFx)
    fprintf("%d, ", bestX);
    fprintf("]\n");
    if r2adj==1, break; end
  end
  % Start implementing the quasi–nested loops
  idx(1) = idx(1) + iStep(1);
  if idx(1) > iTo(1)
    idx(1) = iFrom(1);
    for i = 2:nVars
      idx(i) = idx(i) + iStep(i);
      if idx(i) > iTo(i) && i < nVars
        idx(i) = iFrom(i);
      elseif idx(i) > iTo(i) && i == nVars
        bStop = true;
      else
        break
      end
    end
  end
end
fprintf("Regression model\n")
format long
AdjR2 = bestMdl.Rsquared.Adjusted;
fprintf("Adjusted R-square = %14.12f\n", AdjR2);
mdl = bestMdl
bestX = [AdjR2 bestX];

res = zeros(2,length(bestX));
res(1,:) = bestX;
for i=1:5
  res(2,i+1) = mdl.Coefficients{i,1}';
end
writematrix(res(1,:), filename, 'Sheet', 'Results', 'Range', 'A2:Z2');
writematrix(res(2,2:6), filename, 'Sheet', 'Results', 'Range', 'B3:Z3');
 
% now do the % projections
ycalc = project(res);
percErr = 100.*(ycalc-ydata)./ydata;
ProjMat = [ydata xdata ycalc percErr];
writematrix(ProjMat, filename, 'Sheet', 'Project', 'Range', strcat("A2:Z", num2str(1+length(ydata))));


i1 = res(1,2);
i2 = res(1,3);
i3 = res(1,4);
i4 = res(1,5);
px1 = res(1,6);
px2 = res(1,7);
px3 = res(1,8);
px4 = res(1,9);
orient = res(1,10:13);


if px1==0
  sX1 = sprintf("log(X%d)", i1);
elseif px1==4
  sX1 = sprintf("sqrt(X%d)", i1);
elseif px1==1
  sX1 = sprintf("X%d", i1);
elseif px1>1
  sX1 = sprintf("X%d^%d", i1, px1);
elseif px1<0
  sX1 = sprintf("1/X%d^%d", i1, abs(px1));
else
  sX1 = sprintf("X%d", i1);
end

if px2==0
  sX2 = sprintf("log(X%d)", i2);
elseif px2==4
  sX2 = sprintf("sqrt(X%d)", i2);
elseif px2==1
  sX2 = sprintf("X%d", i2);
elseif px2>1
  sX2 = sprintf("X%d^%d", i2, px2);
elseif px2<0
  sX2 = sprintf("1/X%d^%d", i2, abs(px2));
else
  sX2 = sprintf("X%d", i2);
end  

if px3==0
  sX3 = sprintf("log(X%d)", i3);
elseif px3==4
  sX3 = sprintf("sqrt(X%d)", i3);
elseif px3==1
  sX3 = sprintf("X%d", i3);
elseif px3>1
  sX3 = sprintf("X%d^%d", i3, px3);
elseif px3<0
  sX3 = sprintf("1/X%d^%d", i3, abs(px3));
else
  sX3 = sprintf("X%d", i3);
end  

if px4==0
  sX4 = sprintf("log(X%d)", i4);
elseif px4==4
  sX4 = sprintf("sqrt(X%d)", i4);
elseif px4==1
  sX4 = sprintf("X%d", i4);
elseif px4>1
  sX4 = sprintf("X%d^%d", i4, px4);
elseif px4<0
  sX4 = sprintf("1/X%d^%d", i4, abs(px4));
else
  sX4 = sprintf("X%d", i4);
end  

sNum = sprintf("Y = (%e + (%e)*%s", res(2,2), res(2,3), sX1);
sDenom = "1 + ";
if orient(2)==1
  sNum = sprintf("%s + (%e)*%s", sNum, res(2,4), sX2);
else
  sDenom = sprintf("%s + (%e)*%s", sDenom, res(2,4), sX2);
end

if orient(3)==1
  sNum = sprintf("%s + (%e)*%s", sNum, res(2,5), sX3);
else
  sDenom = sprintf("%s + (%e)*%s", sDenom, res(2,5), sX3);
end
sDenom = sprintf("%s + (%e)*%s", sDenom, res(2,6), sX4);
fprintf("%s)\n/(%s)\n", sNum, sDenom)

format short
fprintf("Done!\n");
diary off
beep on
for i=1:3
  beep;
  pause(1);
end

function x = myfx(x,p)
  if p==0
    x = log(x);
  elseif p==4
    x = sqrt(x);  
  elseif p > 1
    x = x.^p;
  elseif p < 0
    x = 1./x.^abs(p);
  end
end

function r = myFit(coeff)
  global mdl
  global xdata
  global ydata


  y = ydata;
  idx = coeff(1:4);
  px = coeff(5:8);
  orient = coeff(9:12);

  if idx(1)==idx(2) && orient(1) == orient(2)
    r = 0;
    return;
  end

  if idx(2)==idx(3) && orient(2) == orient(3)
    r = 0;
    return;
  end  

  if idx(3)==idx(4) && orient(3) == orient(4)
    r = 0;
    return;
  end  

  x1 = myfx(xdata(:,idx(1)),px(1));
  if orient(2)==1
    x2 = myfx(xdata(:,idx(2)),px(2));
  else
    x2 = -ydata.*myfx(xdata(:,idx(2)),px(2));
  end
  if orient(3)==1
    x3 = myfx(xdata(:,idx(3)),px(3));
  else
    x3 = -ydata.*myfx(xdata(:,idx(3)),px(3));
  end
  x4 = -ydata.*myfx(xdata(:,idx(4)),px(4));
  X=[x1 x2 x3 x4];
  mdl = fitlm(X,y);
  r = mdl.Rsquared.Adjusted;
end

function ycalc = project(res)
  global xdata

  idx = res(1,2:5);
  px = res(1,6:9);
  orient = res(1,10:13);

  x1 = myfx(xdata(:,idx(1)),px(1)); 
  x2 = myfx(xdata(:,idx(2)),px(2)); 
  x3 = myfx(xdata(:,idx(3)),px(3)); 
  x4 = myfx(xdata(:,idx(4)),px(4)); 
  xNum = res(2,2) + res(2,3) * x1;
  xDenom = 1 + res(2,6) * x4;
  if orient(2) == 1
    xNum = xNum + res(2,4) * x2;
  else
    xDenom = xDenom + res(2,4) * x2;
  end
  if orient(3) == 1
    xNum = xNum + res(2,5) * x3;
  else
    xDenom = xDenom + res(2,5) * x3;
  end
  ycalc = xNum ./ xDenom;
end